/* This file is part of the Views Template Library, written by
 * Gary Powell & Martin Weiser
 *
 * Copyright (c) 1999  Gary Powell
 * Copyright (c) 1999  Konrad Zuse Zentrum fuer Informationstechnik Berlin
 *
 * This material is provided "as is", with absolutely no warranty expressed
 * or implied. Any use is at your own risk.
 *
 * Permission to use or copy this software for any purpose is hereby granted 
 * without fee, provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 */

#if !defined(_POLYMORPHIC_VIEW_H)
#define _POLYMORPHIC_VIEW_H 1

#if !defined(_VIEW_TRAITS_H)
#include "view_traits.h"
#endif 

#if !defined(_TRANSFORM_VIEW_H)
#include "transform_view.h" 
#endif

#if !defined(_FUNCTORS_H)
#include "functors.h"
#endif

// forward declaration so we don't have to include <algorithm>
namespace std {

template <class InputIterator, class UnaryFunction>
UnaryFunction for_each(InputIterator first, InputIterator last, UnaryFunction f);

};
  
BEGIN_VIEW_NAMESPACE
//
// A polymorphic view.
//

template <class container,
          class const_tag = const_view_tag,
          template <class C, class tag> class proxy_template = view_ref>
class polymorphic_view
  : public transform_view<container,
                          dereference<const_traits<typename container::value_type>::type>,
                          const_tag,                                            
                          dereference<const_traits<typename container::value_type>::type>::result_type,
                          proxy_template> {
public:
  typedef dereference<const_traits<typename container::value_type>::type> transform_type;
  typedef transform_view<container,transform_type,const_tag,transform_type::result_type,proxy_template> inherited;

  // containers have to be default constructible
  explicit polymorphic_view() {}
  
  explicit polymorphic_view(inherited::ctor_arg_type& cont)
    : inherited(cont,transform_type()) {}
};

//
// A polymorphic view with reference counted shared ownership semantics. At the end of the
// container's lifetime, a functor can be applied to all the elements in the container.
//

template <class container,
          class const_tag = const_view_tag,
          class functor = nop_func<typename container::value_type> >
class polymorphic_shared_view
  : public polymorphic_view<container,const_tag> {
public:
  typedef polymorphic_view<container,const_tag> inherited;
  typedef view_traits<container,const_tag>::container_type domain_type;
  typedef pair<size_t,domain_type> ref_type;
  
  // This is a tricky implementation detail used in a constructor.
  struct initializer {
    ref_type* ref;
    initializer(container const& cont) 
      : ref(new ref_type(0,cont)) {}
  };
  

  // Containers have to be default constructible, even if default constructed
  // views are good for nothing.
  explicit polymorphic_shared_view(): ref(0) {}

  // Construction from an existing container (and possibly a destruction functor).
  explicit polymorphic_shared_view(initializer init, functor const& f_=functor())
    : inherited(init.ref->second), ref(init.ref), f(f_) { ++ref->first; }

  polymorphic_shared_view(polymorphic_shared_view const& view) 
    : inherited(view), ref(view.ref), f(view.f) { 
    if (ref) ++ref->first; // maybe view is just default constructed
  }

  polymorphic_shared_view& operator=(polymorphic_shared_view const& view) {
    if (this!=&view) {
      inherited::operator=(view);
      release();
      ref = view.ref;
      f = view.f;
      if (ref) ++ref->first; // maybe view is just default constructed
    }
    return *this;
  }

  ~polymorphic_shared_view() { release(); }

private:
  ref_type* ref;
  functor   f;

  void release() {
    if (ref && --ref->first==0) {
	  std::for_each(ref->second.begin(),ref->second.end(),f);
      delete ref;
      ref = 0;
    }
  }
};


END_VIEW_NAMESPACE
#endif
// $Id: polymorphic_view.h,v 1.2 2000/04/11 09:09:21 bzfweise Exp $ 
