/* This file is part of the Views Template Library, written by
 * Gary Powell & Martin Weiser
 *
 * Copyright (c) 1999  Gary Powell
 * Copyright (c) 1999  Konrad Zuse Zentrum fuer Informationstechnik Berlin
 *
 * This material is provided "as is", with absolutely no warranty expressed
 * or implied. Any use is at your own risk.
 *
 * Permission to use or copy this software for any purpose is hereby granted 
 * without fee, provided the above notices are retained on all copies.
 * Permission to modify the code and to distribute modified code is granted,
 * provided the above notices are retained, and a notice that the code was
 * modified is included with the above copyright notice.
 */
#if !defined (_ZIP_VIEW_H)
#define _ZIP_VIEW_H	1

#if !defined(_VIEW_TRAITS_H)
#include "view_traits.h"
#endif

#if !defined(_TWO_CONTAINERS_BASE_H)
#include "two_containers_base.h"
#endif

#if !defined (_REFERENCE_PAIR_H)
#include "reference_pair.h"
#endif

#if !defined(_SYS_LIBRARY_ITERATOR_)
#include <iterator>
#define _SYS_LIBRARY_ITERATOR_
#endif

#if !defined(_SYS_LIBRARY_STDEXCEPT_)
#include <stdexcept>
#define _SYS_LIBRARY_STDEXCEPT_
#endif

//
// A zip view iterator template
//


// forward declaration so we don't have to include <algorithm>
namespace std {

template <class _Tp> 
void swap(_Tp&, _Tp&);

template <class _InputIter1, class _InputIter2>
bool lexicographical_compare(_InputIter1, _InputIter1,
                             _InputIter2, _InputIter2);
};

BEGIN_VIEW_NAMESPACE
// Forward declare check_equal so that you don't have to include functors.h
template <class iter_t_1, class iter_t_2>
inline bool check_equal(iter_t_1, iter_t_1,
                  iter_t_2, iter_t_2);



template <class iter_t_a, class iter_t_b,
          template<class A, class B> class combinator,	// ref_pair
          class range_t>	
class zip_iterator {
public:
  typedef member_access_proxy<combinator<iter_t_a,iter_t_b> > ma_type;
  // It seems as if the standard defines value_type to be non-const, whatever the reason
  // may be. At least it violates the requirement that for every iterator i the expression
  // *i must be convertible to value_type.
  // Since value_type now is in general non-const although we get a const one, we have to
  // do some ugly trickery here.
  typedef dereference_traits<typename iterator_traits<iter_t_a>::reference>::type value_type_a;
  typedef dereference_traits<typename iterator_traits<iter_t_b>::reference>::type value_type_b;
  
  typedef combine_iterator_tags<iter_t_a,iter_t_b>::type   iterator_category;
  typedef combinator<value_type_a,value_type_b>            value_type;
  typedef std::iterator_traits<iter_t_a>::difference_type  difference_type;
  typedef value_type                                       reference;
  typedef value_type const                                 const_reference;
  typedef value_type*                                      pointer;

  
  // Iterators must be default constructible.
  explicit zip_iterator() {}

  template<class view>
  zip_iterator(view& v, iter_t_a const &iter_a_, iter_t_b const &iter_b_)
    : range(v), iter_a(iter_a_), iter_b(iter_b_) {}

  // This should be done by the template above.
  //zip_iterator(const range_t& range_, iter_t_a const &iter_a, iter_t_b const &iter_b_)
  //  : range(range_), iter_a(iter_a_), iter_b(iter_b_) {}
  
  value_type const operator*() const { return value_type(*iter_a,*iter_b); }
  value_type       operator*()       { return value_type(*iter_a,*iter_b); }

  ma_type const operator->() const { return ma_type(value_type(*iter_a,*iter_b)); }
  ma_type       operator->()       { return ma_type(value_type(*iter_a,*iter_b)); }

  // Forward stepping.
  zip_iterator& operator++() { ++iter_a; ++iter_b; return *this; }
  zip_iterator operator++(int) { zip_iterator tmp=*this; ++*this; return tmp; }

  // Backward stepping. This works only with bidirectional iterators.
  zip_iterator& operator--() { --iter_b; --iter_a; return *this; }
  zip_iterator operator--(int) { zip_iterator tmp=*this; --*this; return tmp; }

  zip_iterator& operator+=(difference_type n) {
    std::advance(iter_a,n);
    std::advance(iter_b,n);
    return *this; 
  }

  zip_iterator& operator-=(difference_type n) { 
    // Works only with bidirectional iterators!
    std::advance(iter_a,-n);
    std::advance(iter_b,-n);
    return *this; 
  }
  
  difference_type operator-(const zip_iterator& y ) const {
    return y.iter_a<iter_a? std::distance(y.iter_a,iter_a): -std::distance(iter_a,y.iter_a); }

  zip_iterator operator+(difference_type n) const { zip_iterator tmp(*this); return tmp += n; }
  zip_iterator operator-(difference_type n) const { zip_iterator tmp(*this); return tmp -= n; }
  
  
  
  // Comparisons
  template <class iter_t_a_2, class iter_t_b_2, class range_t_2>
  bool operator==(const zip_iterator<iter_t_a_2,iter_t_b_2,combinator,range_t_2>& rhs) const {
    return iter_a==rhs.iter_a && iter_b==rhs.iter_b; }

  template <class iter_t_a_2, class iter_t_b_2, class range_t_2>
  bool operator<(const zip_iterator<iter_t_a_2,iter_t_b_2,combinator,range_t_2>& rhs) const  { 
   return iter_a < rhs.iter_a;  }

  // Conversion from iterator to const_iterator. This works only if
  // there is a conversion from iter_t to const_iter and range_t to const_range_t.
  template <class const_iter_a, class const_iter_b, class const_range_t>
  operator zip_iterator<const_iter_a,const_iter_b,combinator,const_range_t>() const {
    return zip_iterator<const_iter_a,const_iter_b,combinator,const_range_t>(range,iter_a,iter_b); }

protected:
  range_t range;
  iter_t_a iter_a;
  iter_t_b iter_b;

private:

  template <class iter_t_a_2, class iter_t_b_2,
            template<class A, class B> class combinator_2,
            class range_t_2>
  friend class zip_iterator;

};


//
// A zip view.
//
// The iterators for the iter_base template can be const_iterators, since they
// only mark the ranges.
//
template <class container_a, class container_b,
          class const_tag = const_view_tag,
          template<class A, class B> class combinator = reference_pair,
          class iterator_tag = typename combine_iterator_tags<typename container_a::iterator,
							      typename container_b::iterator>::type,
          template<class container, class const_tag> class proxy_template_a = view_ref,
          template<class container, class const_tag> class proxy_template_b = view_ref>
class zip_view {
public:
  typedef combinator<typename view_traits<container_a,const_tag>::value_type,
                     typename view_traits<container_b,const_tag>::value_type> value_type;
  typedef typename container_a::pointer         pointer;
  typedef typename container_a::reference       reference;
  typedef typename container_a::const_reference const_reference;
  typedef typename container_a::size_type       size_type;
  typedef typename container_a::difference_type difference_type;
  
  typedef view_traits<container_a,const_tag>::container_type domain_type_a;
  typedef view_traits<container_b,const_tag>::container_type domain_type_b;
  typedef view_traits<container_a,const_tag>::iterator domain_a_iterator;
  typedef view_traits<container_a,const_tag>::const_iterator domain_a_const_iterator;
  typedef view_traits<container_b,const_tag>::iterator domain_b_iterator;
  typedef view_traits<container_b,const_tag>::const_iterator domain_b_const_iterator;

  typedef zip_view<container_a,container_b,const_tag,combinator,iterator_tag,
                   proxy_template_a,proxy_template_b> self;

  typedef two_containers_base<container_a,container_b,const_tag,proxy_template_a,proxy_template_b> base_t;

  typedef internal_two_range_traits<iterator_tag, container_a, container_b,const_tag>::type	   iter_base;
  typedef internal_two_range_traits<iterator_tag, container_a, container_b,const_tag>::const_type  const_iter_base;

  typedef zip_iterator<domain_a_iterator,domain_b_iterator,combinator,iter_base> iterator;
  typedef zip_iterator<domain_a_const_iterator,domain_b_const_iterator,
                       combinator,const_iter_base> const_iterator;
  typedef std::reverse_iterator<iterator> reverse_iterator;
  typedef std::reverse_iterator<const_iterator> const_reverse_iterator;

  typedef ctor_arg<container_a,const_tag,proxy_template_a>::type ctor_arg_type_a;
  typedef ctor_arg<container_b,const_tag,proxy_template_b>::type ctor_arg_type_b;

  explicit zip_view() {}

  zip_view(ctor_arg_type_a& cont_a, ctor_arg_type_b& cont_b)
    : base(cont_a,cont_b) {}

  // The end() methods should be optimized for speed. (Test for size equality is a first
  // attempt.)
  const_iterator begin() const { return const_iterator(base,base.a().begin(),base.b().begin()); }
  
  const_iterator end() const   {
    size_type sa = base.a().size();
    size_type sb = base.b().size();
    if (sa==sb) return const_iterator(base,base.a().end(),base.b().end());
    else {
      const_iterator i(begin());
      std::advance(i,std::min(sa,sb));
      return i;
    }
  }
  
  iterator       begin()       { return iterator(base,base.a().begin(),base.b().begin()); }
  
  iterator       end()         { 
    size_type sa = base.a().size();
    size_type sb = base.b().size();
    if (sa==sb) return iterator(base,base.a().end(),base.b().end());
    else {
      iterator i(begin());
      std::advance(i,std::min(sa,sb));
      return i;
    }
  }

  const_reverse_iterator rbegin() const { return const_reverse_iterator(end()); }
  const_reverse_iterator rend() const   { return const_reverse_iterator(begin()); }
  reverse_iterator       rbegin()       { return reverse_iterator(end()); }
  reverse_iterator       rend()         { return reverse_iterator(begin()); }

  size_type size() const { return std::min(base.a().size(),base.b().size()); }
  size_type max_size() const { return std::min(base.a().max_size(),base.b().max_size()); }
  bool empty() const { return base.a().empty() || base.b().empty(); }

  value_type operator[](size_type n) const {
    return value_type(base.a()[n],base.b()[m]); }
  
  value_type at(size_type n) const { 
    range_check(n);
    return value_type(base.a()[n],base.b()[m]); }
  
  void swap(zip_view &rhs) { std::swap(base,rhs.base); }


protected:
  base_t base;

private:  
  void range_check(size_type n) const {
    if (n < 0 || n >= size()) throw std::range_error("zip_view");

  }
};


//
// Global comparator operations.
//
template <class container_a_1,class container_b_1,class container_a_2, class container_b_2,
          class const_tag_1, class const_tag_2,
          template<class A, class B> class combinator_1,
          template<class A, class B> class combinator_2,
          class iterator_tag_1, class iterator_tag_2,
          template<class container, class const_tag> class proxy_template_a_1,
          template<class container, class const_tag> class proxy_template_b_1,
          template<class container, class const_tag> class proxy_template_a_2,
          template<class container, class const_tag> class proxy_template_b_2>
bool operator==(zip_view<container_a_1,container_b_1,const_tag_1,combinator_1,
			 iterator_tag_1,proxy_template_a_1,proxy_template_b_1> const& rhs,
                zip_view<container_a_2,container_b_2,const_tag_2,combinator_2,
			 iterator_tag_2,proxy_template_a_2,proxy_template_b_2> const& lhs) {
  return check_equal(lhs.begin(), lhs.end(), rhs.begin(), rhs.end());
}

template <class container_a_1,class container_b_1,class container_a_2, class container_b_2,
          class const_tag_1, class const_tag_2,
          template<class A, class B> class combinator_1,
          template<class A, class B> class combinator_2,
          class iterator_tag_1, class iterator_tag_2,
          template<class container, class const_tag> class proxy_template_a_1,
          template<class container, class const_tag> class proxy_template_b_1,
          template<class container, class const_tag> class proxy_template_a_2,
          template<class container, class const_tag> class proxy_template_b_2>
bool operator<(zip_view<container_a_1,container_b_1,const_tag_1,combinator_1,
			iterator_tag_1,proxy_template_a_1,proxy_template_b_1> const& rhs,
               zip_view<container_a_2,container_b_2,const_tag_2,combinator_2,
			iterator_tag_2,proxy_template_a_2,proxy_template_b_2> const& lhs) {
	using std::lexicographical_compare;
	return lexicographical_compare(rhs.begin(), rhs.end(), lhs.begin(), lhs.end());
}

END_VIEW_NAMESPACE

namespace std {
//
// The swap function
//
template <class container_a,class container_b,
          class const_tag,
          template<class A, class B> class combinator,
          class iterator_tag,
          template<class container, class const_tag> class proxy_template_a,
          template<class container, class const_tag> class proxy_template_b>
void swap(VIEW_NAMESPACE::zip_view<container_a,container_b,const_tag,combinator,
                                   iterator_tag,proxy_template_a,proxy_template_b> const& a,
          VIEW_NAMESPACE::zip_view<container_a,container_b,const_tag,combinator,
                                   iterator_tag,proxy_template_a,proxy_template_b> const& b) {
  a.swap(b);
}

};


#endif	// _ZIP_VIEW_H
